/* ==========================================================================*/
/* - - - - - - - - - - - - - - L C D . C   - - - - - - - - - - - - - - - - - */
/* ==========================================================================*/
/*	Created   06-14-2004   M. Wheatley                                       */
/*	Modified  08-28-2004   MW                                                */
/*.........................................................................  */
/* Module to Initialize and use the Demo Board LCD Module.                   */
/*.........................................................................  */

/*---------------------------------------------------------------------------*/
/*------------------------> I N C L U D E S <--------------------------------*/
/*---------------------------------------------------------------------------*/
#include <p30f6014.h>
#include <spi.h>
#include <dsp.h>
#include "common.h"
#include "timers.h"
#include "procsignal.h"
#include "menu.h"
#include "gettime.h"
#include "lcd.h"

/*---------------------------------------------------------------------------*/
/*----------------> P R O C E S S O R    S P E C I F I C <-------------------*/
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/*--------> T Y P E D E F S   U N I O N S    S T R U C T U R E S <-----------*/
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/*--------------------> L O C A L   D E F I N E S <--------------------------*/
/*---------------------------------------------------------------------------*/
#define MAXCHARSPERLINE 20
/*---------------------------------------------------------------------------*/
/*----------------> E X T E R N A L  R E F E R E N C E S <-------------------*/
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/*--------------> G L O B A L  M E M O R Y  A L L O C A T I O N <------------*/
/*---------------------------------------------------------------------------*/
char PendPos;

/*---------------------------------------------------------------------------*/
/*--------------> L O C A L  M E M O R Y  A L L O C A T I O N <--------------*/
/*---------------------------------------------------------------------------*/
static unsigned char TxQueue[256];	//must be 256 bytes long for ptr wrap around
static volatile unsigned char TxQHead;
static volatile unsigned char TxQTail;
static char PlotBuf[128];
static int PlotPos;
static char PixelSendState;
static char PendDir;
static char CurPendPos;
static char PendTimer;

static const char INPUT_STR[] = "Input Signal";
static const char SYNC_STR[] = "Seconds         Sync";
static const char DATA_STR[] = "Data Bit";
/*---------------------------------------------------------------------------*/
/*--------------------------> P R O T O T Y P E S <--------------------------*/
/*---------------------------------------------------------------------------*/
void InitLcd(void);
void CursState(char state);
void WriteChar(char ch);
void Write4Hex(unsigned int val);
void Write2Hex(unsigned char val);
void Write2Dec(unsigned char val);
void WriteDec(unsigned int val);
void ClearLcd(void);
void ClearLcdLine(unsigned char y);
void WriteCStr(const char* ptr);
void SetLcdPos(unsigned char x, unsigned char y );
void PlotTime( int* pBuf );
void ServiceLCD(void);
void PlotPendulum( void );
static void WritePend(char new);
static void PutTxQ(unsigned char val);

/*---------------------------------------------------------------------------*/
/*----------> I N T E R R U P T    S E R V I C E    R O U T I N E   <--------*/
/*---------------------------------------------------------------------------*/

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                     ----------------------------                          */
/*                    | S P I 2  I n t e r r u p t |                         */
/*                     ----------------------------                          */
/*  Interrupt Service routine for SPI2 transmitter.                          */
/*                max 4.5uSec execution @ 4xPLL                              */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void __attribute__((__interrupt__)) _SPI2Interrupt(void)
{
	if( TxQHead!=TxQTail )
	{
		if( !SPI2STATbits.SPITBF )
		    WriteSPI2( TxQueue[TxQTail++] );
	}
	SPI2STATbits.SPIROV = 0;
	IFS1bits.SPI2IF = 0;
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                             ---------------                               */
/*                            | I n i t L c d |                              */
/*                             ---------------                               */
/*    This initializes the LCD controller.                                   */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void InitLcd(void)
{
	TRISGbits.TRISG9 = 0;	//SS LCD controller Select line
	PORTGbits.RG0 = 1;
	TxQHead = 0;
	TxQTail = 0;
	CloseSPI2();
	ConfigIntSPI2(SPI_INT_EN & SPI_INT_PRI_3);
	OpenSPI2( FRAME_ENABLE_OFF &
	            ENABLE_SDO_PIN &
	            SPI_MODE16_OFF &
		        SPI_SMP_OFF &
		        SPI_CKE_OFF &
		        SLAVE_ENABLE_OFF &
		        CLK_POL_ACTIVE_HIGH &
		        MASTER_ENABLE_ON &
#if (CLOCK_FREQ == CLOCK_FREQ4XPLL)
		        SEC_PRESCAL_1_1 &		//LCD seems to require ~100KHz clk
#else
		        SEC_PRESCAL_2_1 &		//LCD seems to require ~100KHz clk
#endif
		        PRI_PRESCAL_64_1,		//faster clks overrun its buffer?
				SPI_ENABLE &
	            SPI_IDLE_CON &
	            SPI_RX_OVFLOW_CLR
			);
    LATGbits.LATG9 = 0;
	SemA.PlotActive = 0;
	TimeDelay(2);
    WriteSPI2(0x80);	//reset display
	TimeDelay(2);
    WriteSPI2(0x80);	//reset display
	TimeDelay(2);
	PendDir = 1;
	PendPos = 0;
	PendTimer = 0;
}


/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                            -------------                                  */
/*                           | P u t T x Q |                                 */
/*                            -------------                                  */
/*    This places one byte into the SPI transmit Queue.                      */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
static void PutTxQ(unsigned char val)
{
	DisableIntSPI2;
	TxQueue[TxQHead] = val;
	if( TxQTail == TxQHead++ )
	{
		SPI2STATbits.SPIROV = 0;
		IFS1bits.SPI2IF = 1;	//create fake IRQ to "prime the pump"
	}
	EnableIntSPI2;
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                            -----------------                              */
/*                           | C l e a r L c d |                             */
/*                            -----------------                              */
/*    This clears the LCD display and sets the position to 0,0.              */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void ClearLcd(void)
{
	PutTxQ(0x82);
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                            -------------------                            */
/*                           | C u r s S t a t e |                           */
/*                            -------------------                            */
/*    This clears the LCD display and sets the position to 0,0.              */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void CursState(char state)
{
	if(state==0)
	{
		PutTxQ(0x8B);
	}
	else
	{
	    PutTxQ(0x8C);
	    PutTxQ(0xAD);
	    PutTxQ(0x03);
	}
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                            -------------------------                      */
/*                           | C l e a r L c d L i n e |                     */
/*                            -------------------------                      */
/*    This clears the LCD display line y.                                    */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void ClearLcdLine(unsigned char y)
{
    PutTxQ(0xA9);
    PutTxQ(y);
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                            -------------------                            */
/*                           | W r i t e C h a r |                           */
/*                            -------------------                            */
/*    This writes a single char at the current cursor position.              */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void WriteChar(char ch)
{
    PutTxQ(0xA8);
    PutTxQ(ch);
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                            -------------------                            */
/*                           | W r i t e P e n d |                           */
/*                            -------------------                            */
/*    Clears pendPos and places a marker at NewPos.                          */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
static void WritePend(char newpos)
{
    PutTxQ(0xE6);
    PutTxQ(' ');
    PutTxQ(CurPendPos);
    PutTxQ(3);
//
    PutTxQ(0xE6);
	if( (0==newpos) || (19==newpos) )
	    PutTxQ('*');
	else
	    PutTxQ('-');
    PutTxQ(newpos);
    PutTxQ(3);
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                            -------------------                            */
/*                           | W r i t e C S t r |                           */
/*                            -------------------                            */
/*    This writes a constant string at the current cursor position.          */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void WriteCStr(const char* ptr)
{
unsigned int i=0;
char ch;
	do
	{
		ch = ptr[i++];
		if(ch==0)
			break;
		else
			WriteChar(ch);
	}while(i<MAXCHARSPERLINE);
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                            -------------------                            */
/*                           | W r i t e 4 H e x |                           */
/*                            -------------------                            */
/*    This writes int val in hex format to the LCD.                          */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void Write4Hex(unsigned int val)
{
char ch;
	ch = (val>>12);
	if(ch<=9)
		ch += '0';
	else
		ch += 'A' - 10;
	WriteChar(ch);
	ch = (val>>8) & 0x0F;
	if(ch<=9)
		ch += '0';
	else
		ch += 'A' - 10;
	WriteChar(ch);
	ch = (val>>4) & 0x0F;
	if(ch<=9)
		ch += '0';
	else
		ch += 'A' - 10;
	WriteChar(ch);
	ch = val & 0x0F;
	if(ch<=9)
		ch += '0';
	else
		ch += 'A' - 10;
	WriteChar(ch);
}
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                            -------------------                            */
/*                           | W r i t e 2 H e x |                           */
/*                            -------------------                            */
/*    This writes char val in hex format to the LCD.                         */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void Write2Hex(unsigned char val)
{
char ch;
	ch = val>>4;
	if(ch<=9)
		ch += '0';
	else
		ch += 'A' - 10;
	WriteChar(ch);
	ch = val & 0x0F;
	if(ch<=9)
		ch += '0';
	else
		ch += 'A' - 10;
	WriteChar(ch);
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                            -------------------                            */
/*                           | W r i t e 2 D e c |                           */
/*                            -------------------                            */
/*    This writes val in decimal format to the LCD.                          */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void Write2Dec(unsigned char val)
{
char ch;
	ch = val/10;
	val = val - 10*ch;
	ch += '0';
	WriteChar(ch);
	ch = val;
	ch += '0';
	WriteChar(ch);
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                            -----------------                              */
/*                           | W r i t e D e c |                             */
/*                            -----------------                              */
/*    This writes val in decimal format to the LCD.                          */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void WriteDec(unsigned int val)
{
char ch;
unsigned int tmp;
	tmp = val;
	ch = tmp/10000;
	tmp = tmp - 10000*ch;
	ch += '0';
	WriteChar(ch);
	ch = tmp/1000;
	tmp = tmp - 1000*ch;
	ch += '0';
	WriteChar(ch);
	ch = tmp/100;
	tmp = tmp - 100*ch;
	ch += '0';
	WriteChar(ch);
	ch = tmp/10;
	tmp = tmp - 10*ch;
	ch += '0';
	WriteChar(ch);
	ch = tmp;
	ch += '0';
	WriteChar(ch);
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                            -------------------                            */
/*                           | S e t L c d P o s |                           */
/*                            -------------------                            */
/*    This sets the current cursor position.                                 */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void SetLcdPos(unsigned char x, unsigned char y )
{
    PutTxQ(0xC5);
    PutTxQ(x);
    PutTxQ(y);
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                            --------------                                 */
/*                           | PlotPendulum |                                */
/*                            --------------                                 */
/*    This routine implements a linear pendulum on the main display.         */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void PlotPendulum( void )
{
	if(++PendTimer == 5)	//divide 50Hz rate to 10Hz
	{
		PendTimer = 0;
		if(PendPos>19)		//logic to deal with direction changing
			PendDir = -1;
		else				//and modulo 20 character issues
			if(PendPos<0)
				PendDir = +1;
			else
				if(MENU_MAIN == CurrentMenu)
				{
					WritePend(PendPos);
					CurPendPos = PendPos;
				}
		PendPos = CurPendPos+PendDir;
	}
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                            -----------------                              */
/*                           | P l o t T i m e |                             */
/*                            -----------------                              */
/*    This starts a plot of a 128 word data buffer amplitude versus samples. */
/*  It just scales and places in the plot buffer and initiates the LCD plot  */
/* writing state machine													 */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void PlotTime( int* pBuf )
{
int x;
int z;
	if(!SemA.PlotActive)
	{
		if(0==LCDHoldoffCounter)
		{
			if(CurrentMenu==MENU_SIGNAL)
			{
				for(x=0; x<122; x++)
					PlotBuf[x] = (char)(16 + (pBuf[x]>>11) );
		    	PutTxQ(0x82);	//clear screen
				SetLcdPos(0,0);		//draw fixed text
				WriteCStr(INPUT_STR);
				PlotPos = 0;		//x position
				PixelSendState = 0;
				SemA.PlotActive = 1;		//enable LCD writing state machine
				LCDHoldoffCounter = 4;
			}
			else if( SemA.SecRdy)
			{
		    	PutTxQ(0x82);	//clear screen
				if( CurrentMenu==MENU_SYNC)
				{
					z = 150 - (122/2);
					for(x=0; x<122; x++)
					{
						PlotBuf[x] = (char)( (SecSyncBuf[z]>>9) );
						if(++z >= 150)
							z = 0;
					}
					SetLcdPos(0,0);		//draw fixed text
					WriteCStr(SYNC_STR);
				}
				else if( CurrentMenu==MENU_DATA )
				{
					for(x=0; x<122; x++)
						PlotBuf[x] = (char)( (RawDataBuf[x/4]>>7) );
					SetLcdPos(12,0);		//draw fixed text
					WriteCStr(DATA_STR);
				}
				PlotPos = 0;		//x position
				PixelSendState = 0;
				SemA.PlotActive = 1;		//enable LCD writing state machine
				LCDHoldoffCounter = 4;
			}
		}
	}
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                        ---------------------                              */
/*                       | S e r v i c e L C D |                             */
/*                        ---------------------                              */
/*    Called in main loop to manage plotting of points one at a time.        */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void ServiceLCD( )
{
    if( (TxQTail == TxQHead ) && SemA.PlotActive )	//if still points to plot
	{											//and SPI idle
		PixelSendState++;
		if( PixelSendState==1 )
		{
	    	PutTxQ(0xD8);			//pixel ON cmd
		}
		else if( PixelSendState==2 )
		{
	    	PutTxQ(PlotPos );		//pixel x pos write
		}
		else
		{
	    	PutTxQ(PlotBuf[PlotPos++] );
			PixelSendState = 0;			//pixel y pos write
		}
		if(PixelSendState>=3)
			PixelSendState = 0;
		if( PlotPos>=122 )
	    	SemA.PlotActive = 0;	//done plotting points
	}
}
